import validator from '@navikt/fnrvalidator'
import { getYear, parseISO } from 'date-fns'
import i18n from 'i18next'
import {
  FetchedIdentity,
  FetchedRole,
  Identity,
  Role,
  Consent,
  FetchedConsent,
} from 'interfaces'
import { isValidPhoneNumber, parsePhoneNumber } from 'libphonenumber-js'
import parse from 'date-fns/parse'
import { parseInt } from 'lodash'

const validEmailRegex =
  /^(?=.{1,254}$)(?=.{1,64}@)[-!#$%&'*+/0-9=?A-Z^_`a-z{|}~]+(\.[-!#$%&'*+/0-9=?A-Z^_`a-z{|}~]+)*@[A-Za-z0-9]([A-Za-z0-9-]{0,61}[A-Za-z0-9])?(\.[A-Za-z0-9]([A-Za-z0-9-]{0,61}[A-Za-z0-9])?)*$/

export function getCookie(name: string) {
  if (!document.cookie) {
    return null
  }

  const cookies = document.cookie
    .split(';')
    .map((c) => c.trim())
    .filter((c) => c.startsWith(`${name}=`))

  if (cookies.length === 0) {
    return null
  }
  return decodeURIComponent(cookies[0].split('=')[1])
}

export function setCookie(name: string, value: string) {
  document.cookie = `${name}=${value}; path=/`
}

export function deleteCookie(name: string) {
  if (getCookie(name)) {
    setCookie(name, '; expires=Thu, 01 jan 1970 00:00:00 GMT')
  }
}

export function maybeCsrfToken() {
  const csrfToken = getCookie('csrftoken')
  if (!csrfToken) {
    return null
  }
  return {
    'X-CSRFToken': csrfToken,
  }
}

export function submitJsonOpts(method: string, data: object): RequestInit {
  return {
    method,
    headers: {
      'Content-Type': 'application/json',
      ...maybeCsrfToken(),
    },
    body: JSON.stringify(data),
    credentials: 'same-origin',
  }
}

export function fetchJsonOpts(): RequestInit {
  return {
    headers: {
      'Content-Type': 'application/json',
    },
    credentials: 'same-origin',
  }
}

export function isValidFnr(
  data: string | undefined,
  allowEmpty = false
): boolean | string {
  if (!data) {
    if (allowEmpty) {
      return true
    }
    return i18n.t<string>('common:validation.invalidIdNumber').toString()
  }
  const valid = validator.idnr(data as string).status === 'valid'
  if (valid) {
    return true
  }
  // TypeScript complains if toString is not used on the function result
  return i18n.t<string>('common:validation.invalidIdNumber').toString()
}

export function isValidMobilePhoneNumber(
  data: string | undefined
): boolean | string {
  if (!data) {
    return i18n.t<string>('common:validation.invalidMobilePhoneNumber')
  }
  const valid = isValidPhoneNumber(data)
  if (valid) {
    return true
  }
  return i18n.t<string>('common:validation.invalidMobilePhoneNumber')
}

export function isValidEmail(data: string | undefined): boolean | string {
  if (!data) {
    return i18n.t<string>('common:validation.emailRequired')
  }
  if (validEmailRegex.test(data)) {
    return true
  }
  return i18n.t<string>('common:validation.invalidEmail')
}

/**
 * Splits a phone number into a country code and the national number.
 *
 * @param phoneNumber The phone number to split
 */
export function splitPhoneNumber(phoneNumber: string): [string, string] {
  const parsedNumber = parsePhoneNumber(phoneNumber)

  return [
    parsedNumber.countryCallingCode.toString(),
    parsedNumber.nationalNumber.toString(),
  ]
}

export function getLocalized(
  obj: any,
  prefix: string,
  locale: string
): string | null {
  const key = prefix + locale
  return Object.prototype.hasOwnProperty.call(obj, key) ? obj[key] : null
}

export function parseRole(role: FetchedRole): Role {
  return {
    id: role.id,
    name_nb: role.name_nb,
    name_en: role.name_en,
    ou_nb: role.ou_nb,
    ou_en: role.ou_en,
    start_date: role.start_date == null ? undefined : parseISO(role.start_date),
    end_date: parseISO(role.end_date),
    max_days: role.max_days,
    contact_person_unit: role.contact_person_unit,
    comments: role.comments,
    sponsor_name: role.sponsor_name,
    ou_id: role.ou_id
  }
}

export function parseConsent(cons: FetchedConsent): Consent {
  return {
    id: cons.id,
    type: cons.type,
    choice: cons.choice,
    consent_given_at: cons.consent_given_at
      ? parseISO(cons.consent_given_at)
      : null,
  }
}

export function parseIdentity(
  identity: FetchedIdentity | null
): Identity | null {
  if (identity == null) {
    return null
  }
  return {
    id: identity.id,
    type: identity.type,
    value: identity.value,
    verified_at: identity.verified_at ? parseISO(identity.verified_at) : null,
    verified: identity.verified,
    verified_by: identity.verified_by,
    source: identity.source,
  }
}

/**
 * Method to ensure the correct "styled" case of a institution acronym.
 *
 * @param instName An institution acronym in any case
 * @returns The correct "styled" case acronym, or the input string if no match.
 */
export function instNameUpperCaser(instName: String): String {
  switch (instName.toLocaleLowerCase()) {
    case 'uio':
      return 'UiO'
    case 'uib':
      return 'UiB'
    case 'ntnu':
      return 'NTNU'
    default:
      return instName
  }
}

/**
 * Get the name of a role matching chosen language with fallback to nb/en if missing
 * @param role an object of type Role
 * @returns the name matching the language or fallback if missing
 */
export function getRoleName(role: Role) {
  if (i18n.language === 'en') {
    return role.name_en ? role.name_en : role.name_nb
  }
  return role.name_nb ? role.name_nb : role.name_en
}

/**
 * Get the ou name of a role matching chosen language with fallback to nb/en if missing
 * @param role an object of type Role
 * @returns the ou name matching the language or fallback if missing
 */
export function getRoleOuName(role: Role) {
  if (i18n.language === 'en') {
    return role.ou_en ? role.ou_en : role.ou_nb
  }
  return role.ou_nb ? role.ou_nb : role.ou_en
}

/**
 * Note the the input is assumed to be either a D-number or a "fødselsnummer". Other types such as H-numbers are not supported.
 *
 * @param nationalId D-number or "fødselsnummer"
 */
function isDnr(nationalId: string): boolean {
  return parseInt(nationalId.substring(0, 1), 10) >= 4
}

export function isFemaleBasedOnNationalId(nationalId: string): boolean {
  if (isDnr(nationalId)) {
    return parseInt(nationalId.charAt(10), 10) % 2 === 0
  }
  return parseInt(nationalId.charAt(8), 10) % 2 === 0
}

/**
 * The ID-number format is scheduled to change in the future:
 * https://www.skatteetaten.no/deling/opplysninger/folkeregisteropplysninger/pid/
 *
 * Do not attempt a suggestion if we are at the time when the format will change.
 */
function shouldAttemptSuggestion(): boolean {
  const currentYear = getYear(new Date())

  // Keeping it simple, if this code is still running and should be revised.
  // The new numbers are scheduled to appear in 2032
  return currentYear <= 2031
}

export function extractGenderOrBlank(nationalId?: string): string {
  if (
    nationalId == null ||
    nationalId === '' ||
    isValidFnr(nationalId) !== true ||
    !shouldAttemptSuggestion()
  ) {
    return ''
  }

  if (isFemaleBasedOnNationalId(nationalId)) {
    return 'female'
  }
  return 'male'
}

/**
 * Gives a guess of the birthdate with century included.
 *
 * @param dateOfBirth a date on the form ddMMyy
 */
function suggestBirthDate(dateOfBirth: string): Date {
  const currentYear = getYear(new Date())
  const year = dateOfBirth.substring(4, 6)
  const yearAsInt = parseInt(year)
  let century = '20'

  // Check that the year the person is born is not a year in the future,
  // given he is born in the 21th century. Also assuming he is born in
  // the 21th century, check that he is older than 15 years, if not
  // then assume he is born in the 20th century
  if (
    yearAsInt + 2000 > currentYear ||
    !(currentYear - 2000 - yearAsInt > 15)
  ) {
    century = '19'
  }
  return parse(
    dateOfBirth.substring(0, 4) + century + dateOfBirth.substring(4, 6),
    'ddMMyyyy',
    new Date()
  )
}

function extractBirthdateFromFnr(nationalId: string): Date {
  return suggestBirthDate(nationalId.substring(0, 6))
}

function extractBirthdateFromDnumber(nationalId: string): Date {
  return suggestBirthDate(
    (parseInt(nationalId.charAt(0), 10) - 4).toString(10) +
      nationalId.substring(1, 6)
  )
}

export function extractBirthdateFromNationalId(
  nationalId?: string
): Date | null {
  if (
    nationalId == null ||
    nationalId === '' ||
    isValidFnr(nationalId) !== true ||
    !shouldAttemptSuggestion()
  ) {
    return null
  }

  if (isDnr(nationalId)) {
    return extractBirthdateFromDnumber(nationalId)
  }
  return extractBirthdateFromFnr(nationalId)
}
